class AnalytixError(Exception):
    """The base exception class for analytix."""

    pass


class InvalidScopes(AnalytixError):
    pass


class InvalidRequest(AnalytixError):
    pass


class Deprecated(AnalytixError):
    pass


class MissingOptionalComponents(AnalytixError):
    pass


class HTTPError(AnalytixError):
    pass



__productname__ = "analytix"
__version__ = "2.0.0.dev5"
__description__ = "A simple yet powerful API wrapper to make getting analytical information from the YouTube Analytics API easier than ever."
__url__ = "https://github.com/parafoxia/analytix"
__docs__ = "https://analytix.readthedocs.io/en/latest/"
__author__ = "Ethan Henderson"
__license__ = "BSD-3-Clause"
__bugtracker__ = "https://github.com/parafoxia/analytix/issues"

from .youtube.analytics.api import YouTubeAnalytics



YOUTUBE_ANALYTICS_API_SERVICE_NAME = "youtubeAnalytics"
YOUTUBE_ANALYTICS_API_VERSION = "v2"
YOUTUBE_ANALYTICS_SCOPES = (
    "https://www.googleapis.com/auth/yt-analytics.readonly",
    "https://www.googleapis.com/auth/yt-analytics-monetary.readonly",
)



import csv
import datetime as dt
import json
import logging
import os
import time
import typing as t

import requests
from requests_oauthlib import OAuth2Session

from analytix.errors import *
from analytix.iso import CURRENCIES
from analytix.packages import *
from analytix.secrets import TOKEN_STORE, YT_ANALYTICS_TOKEN
from analytix.youtube.analytics import *
from analytix.youtube.analytics import verify

if PANDAS_AVAILABLE:
    import pandas as pd


class YouTubeAnalytics:
    __slots__ = ("_session", "secrets", "project_id", "_token")

    def __init__(self, session, secrets, **kwargs):
        self._session = session
        self.secrets = secrets
        self.project_id = secrets["project_id"]
        self._token = self._get_token()

    def __str__(self):
        return self.project_id

    def __repr__(self):
        return f"<YouTubeAnalytics project_id={self.project_id!r}>"

    @property
    def authorised(self):
        return bool(self._token)

    authorized = authorised

    @classmethod
    def from_file(cls, path, *, scopes="all", **kwargs):
        if not os.path.isfile(path):
            raise FileNotFoundError(
                "you must provide a valid path to a secrets file"
            )

        with open(path, mode="r", encoding="utf-8") as f:
            logging.debug("Secrets file loaded")
            secrets = json.load(f)["installed"]

        scopes = cls._resolve_scopes(scopes)
        session = OAuth2Session(
            secrets["client_id"],
            redirect_uri=secrets["redirect_uris"][0],
            scope=scopes,
            **kwargs,
        )
        return cls(session, secrets)

    @classmethod
    def from_dict(cls, secrets, *, scopes="all", **kwargs):
        scopes = cls._resolve_scopes(scopes)
        session = OAuth2Session(
            secrets["installed"]["client_id"],
            redirect_uri=secrets["redirect_uris"][0],
            scope=scopes,
            **kwargs,
        )
        return cls(session, secrets["installed"])

    @staticmethod
    def _resolve_scopes(scopes):
        if scopes == "all":
            logging.debug(f"Scopes set to {YOUTUBE_ANALYTICS_SCOPES}")
            return YOUTUBE_ANALYTICS_SCOPES

        if not isinstance(scopes, (tuple, list, set)):
            raise InvalidScopes(
                f"expected tuple, list, or set of scopes, got {type(scopes).__name__}"
            )

        for i, scope in enumerate(scopes[:]):
            if not scope.startswith("https://www.googleapis.com/auth/"):
                scopes[i] = "https://www.googleapis.com/auth/" + scope

        diff = set(scopes) - set(YOUTUBE_ANALYTICS_SCOPES)
        if diff:
            raise InvalidScopes(
                "one or more scopes you provided are invalid: "
                + ", ".join(diff)
            )

        logging.debug(f"Scopes set to {scopes}")
        return scopes

    @staticmethod
    def _get_token():
        if not os.path.isfile(TOKEN_STORE / YT_ANALYTICS_TOKEN):
            logging.info("No token found. You will need to authorise")
            return ""

        with open(
            TOKEN_STORE / YT_ANALYTICS_TOKEN, mode="r", encoding="utf-8"
        ) as f:
            data = json.load(f)
        if time.time() > data["expires"]:
            logging.info(
                "Token found, but it has expired. You will need to authorise"
            )
            return ""

        logging.info(
            "Valid token found! analytix will use this, so you don't need to authorise"
        )
        return data["token"]

    def authorise(self, store_token=True, force=False, **kwargs):
        if self._token and not force:
            logging.warning("Client is already authorised! Skipping...")
            return

        url, _ = self._session.authorization_url(
            self.secrets["auth_uri"], **kwargs
        )
        code = input(f"You need to authorise the session: {url}\nCODE > ")
        token = self._session.fetch_token(
            self.secrets["token_uri"],
            code=code,
            client_secret=self.secrets["client_secret"],
        )
        self._token = token["access_token"]
        logging.info("Token retrieved")

        if not store_token:
            logging.info("Not storing token, as instructed")
            return

        os.makedirs(TOKEN_STORE, exist_ok=True)
        with open(
            TOKEN_STORE / YT_ANALYTICS_TOKEN, mode="w", encoding="utf-8"
        ) as f:
            json.dump(
                {"token": self._token, "expires": token["expires_at"]},
                f,
                ensure_ascii=False,
            )
            logging.info(f"Key stored in {TOKEN_STORE / YT_ANALYTICS_TOKEN}")

    authorize = authorise

    def retrieve(
        self, start_date, end_date=dt.date.today(), metrics="all", **kwargs
    ):
        currency = kwargs.pop("currency", "USD")
        dimensions = kwargs.pop("dimensions", ())
        filters = kwargs.pop("filters", {})
        include_historical_data = kwargs.pop("include_historical_data", False)
        max_results = kwargs.pop("max_results", 0)
        sort_by = kwargs.pop("sort_by", ())
        start_index = kwargs.pop("start_index", 1)

        logging.debug("Verifying options...")
        if "7DayTotals" in dimensions or "30DayTotals" in dimensions:
            raise Deprecated(
                "the '7DayTotals' and '30DayTotals' dimensions were deprecated, and can no longer be used"
            )
        if not isinstance(start_date, dt.date):
            raise InvalidRequest(
                f"expected start date as date object, got {type(start_date).__name__}"
            )
        if not isinstance(end_date, dt.date):
            raise InvalidRequest(
                f"expected end date as date object, got {type(end_date).__name__}"
            )
        if end_date <= start_date:
            raise InvalidRequest(
                f"the start date should be earlier than the end date"
            )
        if currency not in CURRENCIES:
            raise InvalidRequest(
                f"expected existing currency as ISO 4217 alpha-3 code, got {currency}"
            )
        if not isinstance(dimensions, (tuple, list, set)):
            raise InvalidRequest(
                f"expected tuple, list, or set of dimensions, got {type(dimensions).__name__}"
            )
        if not isinstance(filters, dict):
            raise InvalidRequest(
                f"expected dict of filters, got {type(filters).__name__}"
            )
        if not isinstance(include_historical_data, bool):
            raise InvalidRequest(
                f"expected bool for 'include_historical_data', got {type(include_historical_data).__name__}"
            )
        if not isinstance(max_results, int):
            raise InvalidRequest(
                f"expected int for 'max_results', got {type(max_results).__name__}"
            )
        if max_results < 0:
            raise InvalidRequest(
                f"the maximum number of results should be no less than 0 (0 for unlimited results)"
            )
        if not isinstance(sort_by, (tuple, list, set)):
            raise InvalidRequest(
                f"expected tuple, list, or set of sorting columns, got {type(sort_by).__name__}"
            )
        if not isinstance(start_index, int):
            raise InvalidRequest(
                f"expected int for 'start_index', got {type(start_index).__name__}"
            )
        if start_index < 1:
            raise InvalidRequest(f"the start index should be no less than 1")

        logging.debug("Determining report type...")
        rtype = verify.rtypes.determine(dimensions, metrics, filters)()
        logging.info(f"Report type determined as: {rtype}")

        if metrics == "all":
            metrics = tuple(rtype.metrics)
        elif not isinstance(metrics, (tuple, list, set)):
            raise InvalidRequest(
                f"expected tuple, list, or set of metrics, got {type(metrics).__name__}"
            )
        logging.debug("Using these metrics: " + ", ".join(metrics))

        logging.debug("Verifying report...")
        rtype.verify(dimensions, metrics, filters, max_results, sort_by)
        logging.debug("Verification complete")

        url = (
            f"https://youtubeanalytics.googleapis.com/{YOUTUBE_ANALYTICS_API_VERSION}/reports"
            "?ids=channel==MINE"
            f"&metrics={','.join(metrics)}"
            f"&startDate={start_date.strftime('%Y-%m-%d')}"
            f"&endDate={end_date.strftime('%Y-%m-%d')}"
            f"&currency={currency}"
            f"&dimensions={','.join(dimensions)}"
            f"&filters={';'.join(f'{k}=={v}' for k, v in filters.items())}"
            f"&includeHistorialChannelData={f'{include_historical_data}'.lower()}"
            f"&maxResults={max_results}"
            f"&sort={','.join(sort_by)}"
            f"&startIndex={start_index}"
        )
        logging.debug(f"URL: {url}")

        if not self._token:
            logging.debug("Authorising...")
            self.authorise()

        with requests.get(
            url, headers={"Authorization": f"Bearer {self._token}"}
        ) as r:
            data = r.json()

        if next(iter(data)) == "error":
            error = data["error"]
            raise HTTPError(f"{error['code']}: {error['message']}")

        logging.info("Creating report...")
        return YouTubeAnalyticsReport(f"{rtype}", data)


class YouTubeAnalyticsReport:
    __slots__ = ("type", "data", "columns", "_ncolumns", "_nrows")

    def __init__(self, type, data):
        self.type = type
        self.data = data
        self.columns = [c["name"] for c in data["columnHeaders"]]
        self._ncolumns = len(self.columns)
        self._nrows = len(data["rows"])

    def __repr__(self):
        return f"<YouTubeAnalyticsReport shape={self.shape!r}>"

    @property
    def shape(self):
        return (self._nrows, self._ncolumns)

    @requires_pandas
    def to_dataframe(self):
        df = pd.DataFrame(self.data["rows"])
        df.columns = self.columns
        if "day" in df.columns:
            df["day"] = pd.to_datetime(df["day"], format="%Y-%m-%d")
        if "month" in df.columns:
            df["month"] = pd.to_datetime(df["month"], format="%Y-%m")
        return df

    def to_json(self, path, *, indent=4):
        if not path.endswith(".json"):
            path += ".json"

        with open(path, mode="w", encoding="utf-8") as f:
            json.dump(self.data, f, indent=indent, ensure_ascii=False)

    def to_csv(self, path, *, delimiter=","):
        if not path.endswith(".csv"):
            path += ".csv"

        with open(path, mode="w", encoding="utf-8") as f:
            writer = csv.writer(f, delimiter=delimiter)
            writer.writerow(self.columns)
            for r in self.data["rows"]:
                writer.writerow(r)



from analytix.iso import COUNTRIES, SUBDIVISIONS


YOUTUBE_ANALYTICS_CORE_DIMENSIONS = (
    "ageGroup",
    "channel",
    "country",
    "day",
    "gender",
    "month",
    "sharingService",
    "uploaderType",
    "video",
)

YOUTUBE_ANALYTICS_CONTENT_OWNER_DIMENSIONS = (
    "claimedStatus",
    "uploaderType",
)

YOUTUBE_ANALYTICS_ALL_DIMENSIONS = (
    "video",
    "playlist",
    "channel",
    "country",
    "province",
    "day",
    "month",
    "insightPlaybackLocationType",
    "insightPlaybackLocationDetail",
    "liveOrOnDemand",
    "subscribedStatus",
    "youtubeProduct",
    "insightTrafficSourceType",
    "insightTrafficSourceDetail",
    "deviceType",
    "operatingSystem",
    "ageGroup",
    "gender",
    "sharingService",
    "elapsedVideoTimeRatio",
    "audienceType",
    "adType",
    "claimedStatus",
    "uploaderType",
)

YOUTUBE_ANALYTICS_VALID_FILTER_OPTIONS = {
    "video": (),
    "playlist": (),
    "channel": (),
    "group": (),
    "country": COUNTRIES,
    "province": SUBDIVISIONS,
    "continent": (
        "002",
        "019",
        "142",
        "150",
        "009",
    ),
    "subContinent": (
        "014",
        "017",
        "015",
        "018",
        "011",
        "029",
        "013",
        "021",
        "005",
        "143",
        "030",
        "034",
        "035",
        "145",
        "151",
        "154",
        "039",
        "155",
        "053",
        "054",
        "057",
        "061",
    ),
    "day": (),
    "month": (),
    "insightPlaybackLocationType": (
        "BROWSE",
        "CHANNEL",
        "EMBEDDED",
        "EXTERNAL_APP",
        "MOBILE",
        "SEARCH",
        "WATCH",
        "YT_OTHER",
    ),
    "insightPlaybackLocationDetail": (),
    "liveOrOnDemand": (
        "LIVE",
        "ON_DEMAND",
    ),
    "subscribedStatus": (
        "SUBSCRIBED",
        "UNSUBSCRIBED",
    ),
    "youtubeProduct": (
        "CORE",
        "GAMING",
        "KIDS",
        "UNKNOWN",
    ),
    "insightTrafficSourceType": (
        "ADVERTISING",
        "ANNOTATION",
        "CAMPAIGN_CARD",
        "END_SCREEN",
        "EXT_URL",
        "NO_LINK_EMBEDDED",
        "NO_LINK_OTHER",
        "NOTIFICATION",
        "PLAYLIST",
        "PROMOTED",
        "RELATED_VIDEO",
        "SHORTS",
        "SUBSCRIBER",
        "YT_CHANNEL",
        "YT_OTHER_PAGE",
        "YT_PLAYLIST_PAGE",
        "YT_SEARCH",
    ),
    "insightTrafficSourceDetail": (
        "ADVERTISING",
        "CAMPAIGN_CARD",
        "END_SCREEN",
        "EXT_URL",
        "NOTIFICATION",
        "RELATED_VIDEO",
        "SUBSCRIBER",
        "YT_CHANNEL",
        "YT_OTHER_PAGE",
        "YT_SEARCH",
    ),
    "deviceType": (
        "DESKTOP",
        "GAME_CONSOLE",
        "MOBILE",
        "TABLET",
        "TV",
        "UNKNOWN_PLATFORM",
    ),
    "operatingSystem": (
        "ANDROID",
        "BADA",
        "BLACKBERRY",
        "CHROMECAST",
        "DOCOMO",
        "FIREFOX",
        "HIPTOP",
        "IOS",
        "KAIOS",
        "LINUX",
        "MACINTOSH",
        "MEEGO",
        "NINTENDO_3DS",
        "OTHER",
        "PLAYSTATION",
        "PLAYSTATION_VITA",
        "REALMEDIA",
        "SMART_TV",
        "SYMBIAN",
        "TIZEN",
        "WEBOS",
        "WII",
        "WINDOWS",
        "WINDOWS_MOBILE",
        "XBOX",
    ),
    "ageGroup": (
        "age13-17",
        "age18-24",
        "age25-34",
        "age35-44",
        "age45-54",
        "age55-64",
        "age65-",
    ),
    "gender": (
        "female",
        "male",
    ),
    "sharingService": (
        "AMEBA",
        "ANDROID_EMAIL",
        "ANDROID_MESSENGER",
        "ANDROID_MMS",
        "BBM",
        "BLOGGER",
        "COPY_PASTE",
        "CYWORLD",
        "DIGG",
        "DROPBOX",
        "EMBED",
        "MAIL",
        "FACEBOOK",
        "FACEBOOK_MESSENGER",
        "FACEBOOK_PAGES",
        "FOTKA",
        "GMAIL",
        "GOO",
        "GOOGLEPLUS",
        "GO_SMS",
        "GROUPME",
        "HANGOUTS",
        "HI5",
        "HTC_MMS",
        "INBOX",
        "IOS_SYSTEM_ACTIVITY_DIALOG",
        "KAKAO_STORY",
        "KAKAO",
        "KIK",
        "LGE_EMAIL",
        "LINE",
        "LINKEDIN",
        "LIVEJOURNAL",
        "MENEAME",
        "MIXI",
        "MOTOROLA_MESSAGING",
        "MYSPACE",
        "NAVER",
        "NEARBY_SHARE",
        "NUJIJ",
        "ODNOKLASSNIKI",
        "OTHER",
        "PINTEREST",
        "RAKUTEN",
        "REDDIT",
        "SKYPE",
        "SKYBLOG",
        "SONY_CONVERSATIONS",
        "STUMBLEUPON",
        "TELEGRAM",
        "TEXT_MESSAGE",
        "TUENTI",
        "TUMBLR",
        "TWITTER",
        "UNKNOWN",
        "VERIZON_MMS",
        "VIBER",
        "VKONTATKE",
        "WECHAT",
        "WEIBO",
        "WHATS_APP",
        "WYKOP",
        "YAHOO",
        "YOUTUBE_GAMING",
        "YOUTUBE_KIDS",
        "YOUTUBE_MUSIC",
        "YOUTUBE_TV",
    ),
    "elapsedVideoTimeRatio": tuple(f"{n/100}" for n in range(1, 101)),
    "audienceType": (
        "ORGANIC",
        "AD_INSTREAM",
        "AD_INDISPLAY",
    ),
    "adType": (
        "auctionBumperInstream",
        "auctionDisplay",
        "auctionInstream",
        "auctionTrueviewInslate",
        "auctionTrueviewInstream",
        "auctionUnknown",
        "reservedBumperInstream",
        "reservedClickToPlay",
        "reservedDisplay",
        "reservedInstream",
        "reservedInstreamSelect",
        "reservedMasthead",
        "reservedUnknown",
        "unknown",
    ),
    "claimedStatus": ("claimed",),
    "uploaderType": (
        "self",
        "thirdParty",
    ),
}

YOUTUBE_ANALYTICS_ALL_FILTERS = tuple(
    YOUTUBE_ANALYTICS_VALID_FILTER_OPTIONS.keys()
)

YOUTUBE_ANALYTICS_CORE_METRICS = (
    "annotationClickThroughRate",
    "annotationCloseRate",
    "averageViewDuration",
    "comments",
    "dislikes",
    "estimatedMinutesWatched",
    "estimatedRevenue",
    "likes",
    "shares",
    "subscribersGained",
    "subscribersLost",
    "viewerPercentage",
    "views",
)

YOUTUBE_ANALYTICS_ALL_METRICS = (
    "views",
    "redViews",
    "comments",
    "likes",
    "dislikes",
    "videosAddedToPlaylists",
    "videosRemovedFromPlaylists",
    "shares",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "averageViewPercentage",
    "annotationClickThroughRate",
    "annotationCloseRate",
    "annotationImpressions",
    "annotationClickableImpressions",
    "annotationClosableImpressions",
    "annotationClicks",
    "annotationCloses",
    "cardClickRate",
    "cardTeaserClickRate",
    "cardImpressions",
    "cardTeaserImpressions",
    "cardClicks",
    "cardTeaserClicks",
    "subscribersGained",
    "subscribersLost",
    "estimatedRevenue",
    "estimatedAdRevenue",
    "grossRevenue",
    "estimatedRedPartnerRevenue",
    "monetizedPlaybacks",
    "playbackBasedCpm",
    "adImpressions",
    "cpm",
)

YOUTUBE_ANALYTICS_ALL_PROVINCE_METRICS = (
    "views",
    "redViews",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "averageViewPercentage",
    "annotationClickThroughRate",
    "annotationCloseRate",
    "annotationImpressions",
    "annotationClickableImpressions",
    "annotationClosableImpressions",
    "annotationClicks",
    "annotationCloses",
    "cardClickRate",
    "cardTeaserClickRate",
    "cardImpressions",
    "cardTeaserImpressions",
    "cardClicks",
    "cardTeaserClicks",
)

YOUTUBE_ANALYTICS_SUBSCRIPTION_METRICS = (
    "views",
    "redViews",
    "likes",
    "dislikes",
    "videosAddedToPlaylists",
    "videosRemovedFromPlaylists",
    "shares",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "averageViewPercentage",
    "annotationClickThroughRate",
    "annotationCloseRate",
    "annotationImpressions",
    "annotationClickableImpressions",
    "annotationClosableImpressions",
    "annotationClicks",
    "annotationCloses",
    "cardClickRate",
    "cardTeaserClickRate",
    "cardImpressions",
    "cardTeaserImpressions",
    "cardClicks",
    "cardTeaserClicks",
)

YOUTUBE_ANALYTICS_LIVE_PLAYBACK_DETAIL_METRICS = (
    "views",
    "redViews",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
)

YOUTUBE_ANALYTICS_VIEW_PERCENTAGE_PLAYBACK_DETAIL_METRICS = (
    "views",
    "redViews",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "averageViewPercentage",
)

YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS = (
    "views",
    "estimatedMinutesWatched",
)

YOUTUBE_ANALYTICS_ALL_PLAYLIST_METRICS = (
    "views",
    "redViews",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "playlistStarts",
    "viewsPerPlaylistStart",
    "averageTimeInPlaylist",
)

YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_PLAYLIST_METRICS = (
    "views",
    "estimatedMinutesWatched",
    "playlistStarts",
    "viewsPerPlaylistStart",
    "averageTimeInPlaylist",
)



__all__ = (
    "Dimensions",
    "Filters",
    "Metrics",
    "Required",
    "ExactlyOne",
    "OneOrMore",
    "Optional",
    "ZeroOrOne",
    "ZeroOrMore",
)

import abc

from analytix.errors import InvalidRequest

from .constants import *


class FeatureType(metaclass=abc.ABCMeta):
    __slots__ = ("sets",)

    def __init__(self, *sets):
        self.sets = sets
        self.every = []
        for s in sets:
            self.every.extend(s.values)

    def __iter__(self):
        return iter(self.every)

    def verify(self, against):
        raise NotImplementedError(
            "you should not attempt to verify this ABC directly"
        )


class FeatureSet(metaclass=abc.ABCMeta):
    __slots__ = ("values",)

    def __init__(self, *values):
        self.values = set(values)

    def __iter__(self):
        return iter(self.values)

    def verify(self, against, ftype):
        raise NotImplementedError(
            "you should not attempt to verify this ABC directly"
        )


class Dimensions(FeatureType):
    def verify(self, against):
        against = set(against)

        diff = against - set(YOUTUBE_ANALYTICS_ALL_DIMENSIONS)
        if diff:
            raise InvalidRequest(
                f"one or more dimensions you provided are invalid ({', '.join(diff)})"
            )

        diff = against - set(self.every)
        if diff:
            raise InvalidRequest(
                "one or more dimensions you provided are not supported by the "
                f"selected report type ({', '.join(diff)})"
            )

        for s in self.sets:
            s.verify(against, "dimension")


class Filters(FeatureType):
    def __init__(self, *sets):
        self.sets = sets
        self.every = []
        for s in sets:
            self.every.extend([v.split("==")[0] for v in s.values])

    def verify(self, against):
        keys = set(against.keys())

        diff = keys - set(YOUTUBE_ANALYTICS_ALL_FILTERS)
        if diff:
            raise InvalidRequest(
                f"one or more filters you provided are invalid ({', '.join(diff)})"
            )

        diff = keys - set(self.every)
        if diff:
            raise InvalidRequest(
                "one or more filters you provided are not supported by the "
                f"selected report type ({', '.join(diff)})"
            )

        for s in self.sets:
            s.verify(against, "filter")

        for k, v in against.items():
            if (
                against[k]
                and v not in YOUTUBE_ANALYTICS_VALID_FILTER_OPTIONS[k]
            ):
                raise InvalidRequest(
                    f"'{v}' is not a valid value for filter '{k}'"
                )


class Metrics:
    def __init__(self, *values):
        self.values = values

    def __iter__(self):
        return iter(self.values)

    def verify(self, against):
        against = set(against)

        diff = against - set(YOUTUBE_ANALYTICS_ALL_METRICS)
        if diff:
            raise InvalidRequest(
                f"one or more metrics you provided are invalid ({', '.join(diff)})"
            )

        diff = against - set(self.values)
        if diff:
            raise InvalidRequest(
                "one or more metrics you provided are not supported by the "
                f"selected report type ({', '.join(diff)})"
            )


class Required(FeatureSet):
    def verify(self, against, ftype):
        values = []
        for v in self.values:
            k = v.split("==")
            if len(k) == 2:
                if k[0] in against and (k[1] != against[k[0]]):
                    raise InvalidRequest(
                        f"filter '{k[0]}' must be set to '{k[1]}' for the selected report type"
                    )
            values.append(k[0])

        if len(set(values) & set(against)) != len(self.values):
            raise InvalidRequest(
                f"expected all {ftype}s from '{', '.join(values)}', got {len(against)}"
            )


class ExactlyOne(FeatureSet):
    def verify(self, against, ftype):
        if len(self.values & set(against)) != 1:
            raise InvalidRequest(
                f"expected 1 {ftype} from '{', '.join(self.values)}', got {len(against)}"
            )


class OneOrMore(FeatureSet):
    def verify(self, against, ftype):
        if len(self.values & set(against)) == 0:
            raise InvalidRequest(
                f"expected at least 1 {ftype} from '{', '.join(self.values)}', got 0"
            )


class Optional(FeatureSet):
    def verify(self, against, ftype):
        # This doesn't need any verification.
        pass


class ZeroOrOne(FeatureSet):
    def verify(self, against, ftype):
        if len(self.values & set(against)) > 1:
            raise InvalidRequest(
                f"expected 0 or 1 {ftype}s from '{', '.join(self.values)}', got {len(against)}"
            )


class ZeroOrMore(FeatureSet):
    def verify(self, against, ftype):
        # This doesn't need any verification.
        pass



from analytix.errors import InvalidRequest

from .features import *
from .constants import *

_D = Dimensions
_F = Filters
_M = Metrics


class ReportType:
    _friendly_name = "Generic"

    def __init__(self):
        self.dimensions = _D()
        self.metrics = _M()
        self.filters = _F()

    def __str__(self):
        return self._friendly_name

    def verify(self, dim, met, fil, *args):
        self.dimensions.verify(dim)
        self.metrics.verify(met)
        self.filters.verify(fil)


class DetailedReportType(ReportType):
    def __init__(self):
        super().__init__()
        self.max_results = 0

    def verify(self, dim, met, fil, max, srt):
        super().verify(dim, met, fil)

        if not max or max > self.max_results:
            raise InvalidRequest(
                "the 'max_results' parameter must be provided and no larger "
                f"than {self.max_results} for the selected report type"
            )

        if not srt:
            raise InvalidRequest(
                f"you must provide at least 1 sort parameter for the selected report type"
            )

        if any(s.strip("-") not in met for s in srt):
            raise InvalidRequest(
                f"the sort parameter must be one or more valid metrics"
            )

        if any(not s[0] == "-" for s in srt):
            raise InvalidRequest(
                (
                    "you can only sort in descending order for this report type. "
                    "You can do this by prefixing the sort metrics with '-'"
                )
            )


class BasicUserActivity(ReportType):
    _friendly_name = "Basic user activity"

    def __init__(self):
        self.dimensions = _D()
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
        )


class BasicUserActivityUS(ReportType):
    _friendly_name = "Basic user activity (US)"

    def __init__(self):
        self.dimensions = _D()
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_PROVINCE_METRICS)
        self.filters = _F(
            Required("province"),
            ZeroOrOne("video", "group"),
        )


class TimeBasedActivity(ReportType):
    _friendly_name = "Time-based activity"

    def __init__(self):
        self.dimensions = _D(ExactlyOne("day", "month"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
        )


class TimeBasedActivityUS(ReportType):
    _friendly_name = "Time-based activity (US)"

    def __init__(self):
        self.dimensions = _D(ExactlyOne("day", "month"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_PROVINCE_METRICS)
        self.filters = _F(
            Required("province"),
            ZeroOrOne("video", "group"),
        )


class GeographyBasedActivity(ReportType):
    _friendly_name = "Geography-based activity"

    def __init__(self):
        self.dimensions = _D(Required("country"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_METRICS)
        self.filters = _F(
            ZeroOrOne("continent", "subContinent"),
            ZeroOrOne("video", "group"),
        )


class GeographyBasedActivityUS(ReportType):
    _friendly_name = "Geography-based activity (US)"

    def __init__(self):
        self.dimensions = _D(Required("province"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_PROVINCE_METRICS)
        self.filters = _F(
            Required("country==US"),
            ZeroOrOne("video", "group"),
        )


class PlaybackDetailsSubscribedStatus(ReportType):
    _friendly_name = "User activity by subscribed status"

    def __init__(self):
        self.dimensions = _D(
            Optional("subscribedStatus"),
            ZeroOrOne("day", "month"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_SUBSCRIPTION_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            Optional("subscribedStatus"),
        )


class PlaybackDetailsSubscribedStatusUS(ReportType):
    _friendly_name = "User activity by subscribed status (US)"

    def __init__(self):
        self.dimensions = _D(
            Optional("subscribedStatus"),
            ZeroOrOne("day", "month"),
        )
        self.metrics = _M(
            "views",
            "redViews",
            "estimatedMinutesWatched",
            "estimatedRedMinutesWatched",
            "averageViewDuration",
            "averageViewPercentage",
            "annotationClickThroughRate",
            "annotationCloseRate",
            "annotationImpressions",
            "annotationClickableImpressions",
            "annotationClosableImpressions",
            "annotationClicks",
            "annotationCloses",
            "cardClickRate",
            "cardTeaserClickRate",
            "cardImpressions",
            "cardTeaserImpressions",
            "cardClicks",
            "cardTeaserClicks",
        )
        self.filters = _F(
            ZeroOrOne("video", "group"),
            ZeroOrMore("province", "subscribedStatus"),
        )


class PlaybackDetailsLiveTimeBased(ReportType):
    _friendly_name = "Time-based playback details (live)"

    def __init__(self):
        self.dimensions = _D(
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
            ZeroOrOne("day", "month"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LIVE_PLAYBACK_DETAIL_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )


class PlaybackDetailsViewPercentageTimeBased(ReportType):
    _friendly_name = "Time-based playback details (view percentage)"

    def __init__(self):
        self.dimensions = _D(
            Required("country"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LIVE_PLAYBACK_DETAIL_METRICS)
        self.filters = _F(
            ZeroOrOne("continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )


class PlaybackDetailsLiveGeographyBased(ReportType):
    _friendly_name = "Geography-based playback details (live)"

    def __init__(self):
        self.dimensions = _D(
            Required("country"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LIVE_PLAYBACK_DETAIL_METRICS)
        self.filters = _F(
            ZeroOrOne("continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )


class PlaybackDetailsViewPercentageGeographyBased(ReportType):
    _friendly_name = "Geography-based playback details (view percentage)"

    def __init__(self):
        self.dimensions = _D(
            Required("country"),
            ZeroOrMore("subscribedStatus", "youtubeProduct"),
        )
        self.metrics = _M(
            *YOUTUBE_ANALYTICS_VIEW_PERCENTAGE_PLAYBACK_DETAIL_METRICS
        )
        self.filters = _F(
            ZeroOrOne("continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("subscribedStatus", "youtubeProduct"),
        )


class PlaybackDetailsLiveGeographyBasedUS(ReportType):
    _friendly_name = "Geography-based playback details (live, US)"

    def __init__(self):
        self.dimensions = _D(
            Required("province"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LIVE_PLAYBACK_DETAIL_METRICS)
        self.filters = _F(
            Required("country==US"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )


class PlaybackDetailsViewPercentageGeographyBasedUS(ReportType):
    _friendly_name = "Geography-based playback details (view percentage, US)"

    def __init__(self):
        self.dimensions = _D(
            Required("province"),
            ZeroOrMore("subscribedStatus", "youtubeProduct"),
        )
        self.metrics = _M(
            *YOUTUBE_ANALYTICS_VIEW_PERCENTAGE_PLAYBACK_DETAIL_METRICS
        )
        self.filters = _F(
            Required("country==US"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("subscribedStatus", "youtubeProduct"),
        )


class PlaybackLocation(ReportType):
    _friendly_name = "Playback locations"

    def __init__(self):
        self.dimensions = _D(
            Required("insightPlaybackLocationType"),
            ZeroOrMore("day", "liveOrOnDemand", "subscribedStatus"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus"),
        )


class PlaybackLocationDetail(DetailedReportType):
    _friendly_name = "Playback locations (detailed)"

    def __init__(self):
        self.dimensions = _D(Required("insightPlaybackLocationDetail"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            Required("insightPlaybackLocationType==EMBEDDED"),
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus"),
        )
        self.max_results = 25


class TrafficSource(ReportType):
    _friendly_name = "Traffic sources"

    def __init__(self):
        self.dimensions = _D(
            Required("insightTrafficSourceType"),
            ZeroOrMore("day", "liveOrOnDemand", "subscribedStatus"),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus"),
        )


class TrafficSourceDetail(DetailedReportType):
    _friendly_name = "Traffic sources (detailed)"

    def __init__(self):
        self.dimensions = _D(Required("insightTrafficSourceDetail"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            Required("insightTrafficSourceType"),
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus"),
        )
        self.max_results = 25

        # TODO: Need a custom verifier here


class DeviceType(ReportType):
    _friendly_name = "Device types"

    def __init__(self):
        self.dimensions = _D(
            Required("deviceType"),
            ZeroOrMore(
                "day", "liveOrOnDemand", "subscribedStatus", "youtubeProduct"
            ),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore(
                "operatingSystem",
                "liveOrOnDemand",
                "subscribedStatus",
                "youtubeProduct",
            ),
        )


class OperatingSystem(ReportType):
    _friendly_name = "Operating systems"

    def __init__(self):
        self.dimensions = _D(
            Required("operatingSystem"),
            ZeroOrMore(
                "day", "liveOrOnDemand", "subscribedStatus", "youtubeProduct"
            ),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore(
                "deviceType",
                "liveOrOnDemand",
                "subscribedStatus",
                "youtubeProduct",
            ),
        )


class DeviceTypeAndOperatingSystem(ReportType):
    _friendly_name = "Device types and operating systems"

    def __init__(self):
        self.dimensions = _D(
            Required("deviceType", "operatingSystem"),
            ZeroOrMore(
                "day", "liveOrOnDemand", "subscribedStatus", "youtubeProduct"
            ),
        )
        self.metrics = _M(*YOUTUBE_ANALYTICS_LOCATION_AND_TRAFFIC_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus", "youtubeProduct"),
        )


class ViewerDemographics(ReportType):
    _friendly_name = "Viewer demographics"

    def __init__(self):
        self.dimensions = _D(
            OneOrMore("ageGroup", "gender"),
            ZeroOrMore("liveOrOnDemand", "subscribedStatus"),
        )
        self.metrics = _M("viewerPercentage")
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore(
                "deviceType",
                "liveOrOnDemand",
                "subscribedStatus",
                "youtubeProduct",
            ),
        )


class EngagementAndContentSharing(ReportType):
    _friendly_name = "Engagement and content sharing"

    def __init__(self):
        self.dimensions = _D(
            Required("sharingService"),
            Optional("subscribedStatus"),
        )
        self.metrics = _M("shares")
        self.filters = _F(
            ZeroOrOne("country", "continent", "subContinent"),
            ZeroOrOne("video", "group"),
            ZeroOrMore("subscribedStatus"),
        )


class AudienceRetention(ReportType):
    _friendly_name = "Audience retention"

    def __init__(self):
        self.dimensions = _D(Required("elapsedVideoTimeRatio"))
        self.metrics = _M("audienceWatchRatio", "relativeRetentionPerformance")
        self.filters = _F(
            Required("video"),
            ZeroOrMore("audienceType", "subscribedStatus", "youtubeProduct"),
        )

        # TODO: Custom verify


class TopVideosRegional(DetailedReportType):
    _friendly_name = "Top videos by region"

    def __init__(self):
        self.dimensions = _D(Required("video"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_METRICS)
        self.filters = _F(ZeroOrOne("country", "continent", "subContinent"))
        self.max_results = 200


class TopVideosUS(DetailedReportType):
    _friendly_name = "Top videos by state"

    def __init__(self):
        self.dimensions = _D(Required("video"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_ALL_PROVINCE_METRICS)
        self.filters = _F(
            Required("province"),
            Optional("subscribedStatus"),
        )
        self.max_results = 200


class TopVideosSubscribed(DetailedReportType):
    _friendly_name = "Top videos by subscription status"

    def __init__(self):
        self.dimensions = _D(Required("video"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_SUBSCRIPTION_METRICS)
        self.filters = _F(
            Optional("subscribedStatus"),
            ZeroOrOne("country", "continent", "subContinent"),
        )
        self.max_results = 200


class TopVideosYouTubeProduct(DetailedReportType):
    _friendly_name = "Top videos by YouTube product"

    def __init__(self):
        self.dimensions = _D(Required("video"))
        self.metrics = _M(
            *YOUTUBE_ANALYTICS_VIEW_PERCENTAGE_PLAYBACK_DETAIL_METRICS
        )
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrMore("subscribedStatus", "youtubeProduct"),
        )
        self.max_results = 200


class TopVideosPlaybackDetail(DetailedReportType):
    _friendly_name = "Top videos by playback detail"

    def __init__(self):
        self.dimensions = _D(Required("video"))
        self.metrics = _M(*YOUTUBE_ANALYTICS_LIVE_PLAYBACK_DETAIL_METRICS)
        self.filters = _F(
            ZeroOrOne("country", "province", "continent", "subContinent"),
            ZeroOrMore("subscribedStatus", "youtubeProduct"),
        )
        self.max_results = 200


def determine(dimensions, metrics, filters):
    curated = filters.get("isCurated", "0") == "1"

    if "sharingService" in dimensions:
        return EngagementAndContentSharing

    if "elapsedVideoTimeRatio" in dimensions:
        return AudienceRetention

    if "playlist" in dimensions:
        return TopPlaylists

    if "insightPlaybackLocationType" in dimensions:
        if curated:
            return PlaybackLocationPlaylist
        return PlaybackLocation

    if "insightPlaybackLocationDetail" in dimensions:
        if curated:
            return PlaybackLocationDetailPlaylist
        return PlaybackLocationDetail

    if "insightTrafficSourceType" in dimensions:
        if curated:
            return TrafficSourcePlaylist
        return TrafficSource

    if "insightTrafficSourceDetail" in dimensions:
        if curated:
            return TrafficSourceDetailPlaylist
        return TrafficSourceDetail

    if "ageGroup" in dimensions or "gender" in dimensions:
        if curated:
            return ViewerDemographicsPlaylist
        return ViewerDemographics

    if "deviceType" in dimensions:
        if "operatingSystem" in dimensions:
            if curated:
                return DeviceTypeAndOperatingSystemPlaylist
            return DeviceTypeAndOperatingSystem
        if curated:
            return DeviceTypePlaylist
        return DeviceType

    if "operatingSystem" in dimensions:
        if curated:
            return OperatingSystemPlaylist
        return OperatingSystem

    if "video" in dimensions:
        if "province" in filters:
            return TopVideosUS
        if "subscribedStatus" not in filters:
            return TopVideosRegional
        if "province" not in filters and "youtubeProduct" not in filters:
            return TopVideosSubscribed
        if "averageViewPercentage" in metrics:
            return TopVideosYouTubeProduct
        return TopVideosPlaybackDetail

    if "country" in dimensions:
        if "liveOrOnDemand" in dimensions or "liveOrOnDemand" in filters:
            return PlaybackDetailsLiveGeographyBased
        if curated:
            return GeographyBasedActivityPlaylist
        if (
            "subscribedStatus" in dimensions
            or "subscribedStatus" in filters
            or "youtubeProduct" in dimensions
            or "youtubeProduct" in filters
        ):
            return PlaybackDetailsViewPercentageGeographyBased
        return GeographyBasedActivity

    if "province" in dimensions:
        if "liveOrOnDemand" in dimensions or "liveOrOnDemand" in filters:
            return PlaybackDetailsLiveGeographyBasedUS
        if curated:
            return GeographyBasedActivityPlaylistUS
        if (
            "subscribedStatus" in dimensions
            or "subscribedStatus" in filters
            or "youtubeProduct" in dimensions
            or "youtubeProduct" in filters
        ):
            return PlaybackDetailsViewPercentageGeographyBasedUS
        return GeographyBasedActivityUS

    if "youtubeProduct" in dimensions or "youtubeProduct" in filters:
        if "liveOrOnDemand" in dimensions or "liveOrOnDemand" in filters:
            return PlaybackDetailsLiveTimeBased
        return PlaybackDetailsViewPercentageTimeBased

    if "liveOrOnDemand" in dimensions or "liveOrOnDemand" in filters:
        return PlaybackDetailsLiveTimeBased

    if "subscribedStatus" in dimensions:
        if "province" in filters:
            return PlaybackDetailsSubscribedStatusUS
        return PlaybackDetailsSubscribedStatus

    if "day" in dimensions or "month" in dimensions:
        if curated:
            return TimeBasedActivityPlaylist
        if "province" in filters:
            return TimeBasedActivityUS
        return TimeBasedActivity

    if curated:
        return BasicUserActivityPlaylist
    if "province" in filters:
        return BasicUserActivityUS
    return BasicUserActivity


_ALL_REPORTS = {
    "BASIC_USER_ACTIVITY": BasicUserActivity,
    "BASIC_USER_ACTIVITY_US": BasicUserActivityUS,
    "TIME_BASED_ACTIVITY": TimeBasedActivity,
    "TIME_BASED_ACTIVITY_US": TimeBasedActivityUS,
    "GEOGRAPHY_BASED_ACTIVITY": object(),
    "GEOGRAPHY_BASED_ACTIVITY_US": object(),
    "PLAYBACK_DETAILS_SUBSCRIBED_STATUS": object(),
    "PLAYBACK_DETAILS_SUBSCRIBED_STATUS_US": object(),
    "PLAYBACK_DETAILS_LIVE_TIME_BASED": object(),
    "PLAYBACK_DETAILS_VIEW_PERCENTAGE_TIME_BASED": object(),
    "PLAYBACK_DETAILS_LIVE_GEOGRAPHY_BASED": object(),
    "PLAYBACK_DETAILS_VIEW_PERCENTAGE_GEOGRAPHY_BASED": object(),
    "PLAYBACK_DETAILS_LIVE_GEOGRAPHY_BASED_US": object(),
    "PLAYBACK_DETAILS_VIEW_PERCENTAGE_GEOGRAPHY_BASED_US": object(),
    "PLAYBACK_LOCATION": object(),
    "PLAYBACK_LOCATION_DETAIL": object(),
    "TRAFFIC_SOURCE": object(),
    "TRAFFIC_SOURCE_DETAIL": object(),
    "DEVICE_TYPE": object(),
    "OPERATING_SYSTEM": object(),
    "DEVICE_TYPE_AND_OPERATING_SYSTEM": object(),
    "VIEWER_DEMOGRAPHICS": object(),
    "ENGAGEMENT_AND_CONTENT_SHARING": object(),
    "AUDIENCE_RETENTION": object(),
    "TOP_VIDEOS_REGIONAL": object(),
    "TOP_VIDEOS_US": object(),
    "TOP_VIDEOS_SUBSCRIBED": object(),
    "TOP_VIDEOS_YOUTUBE_PRODUCT": object(),
    "TOP_VIDEOS_PLAYBACK_DETAIL": object(),
    "BASIC_USER_ACTIVITY_PLAYLIST": object(),
    "TIME_BASED_ACTIVITY_PLAYLIST": object(),
    "GEOGRAPHY_BASED_ACTIVITY_PLAYLIST": object(),
    "GEOGRAPHY_BASED_ACTIVITY_US_PLAYLIST": object(),
    "PLAYBACK_LOCATION_PLAYLIST": object(),
    "PLAYBACK_LOCATION_DETAIL_PLAYLIST": object(),
    "TRAFFIC_SOURCE_PLAYLIST": object(),
    "TRAFFIC_SOURCE_DETAIL_PLAYLIST": object(),
    "DEVICE_TYPE_PLAYLIST": object(),
    "OPERATING_SYSTEM_PLAYLIST": object(),
    "DEVICE_TYPE_AND_OPERATING_SYSTEM_PLAYLIST": object(),
    "VIEWER_DEMOGRAPHICS_PLAYLIST": object(),
    "TOP_PLAYLISTS": object(),
    "AD_PERFORMANCE": object(),
}



__all__ = (
    "PANDAS_AVAILABLE",
    "requires_pandas",
)

from pkg_resources import working_set

from analytix.errors import MissingOptionalComponents

_packages = [p.key for p in working_set]
PANDAS_AVAILABLE = "pandas" in _packages


def requires_pandas(func):
    def wrapper(*args, **kwargs):
        if not PANDAS_AVAILABLE:
            raise MissingOptionalComponents("pandas is not installed")

        return func(*args, **kwargs)

    return wrapper
